"""
Serialization type ids
"""
# Serialization Constants
CONSTANT_TYPE_NULL = 0
CONSTANT_TYPE_PORTABLE = -1
CONSTANT_TYPE_DATA_SERIALIZABLE = -2
CONSTANT_TYPE_BYTE = -3
CONSTANT_TYPE_BOOLEAN = -4
CONSTANT_TYPE_CHAR = -5
CONSTANT_TYPE_SHORT = -6
CONSTANT_TYPE_INTEGER = -7
CONSTANT_TYPE_LONG = -8
CONSTANT_TYPE_FLOAT = -9
CONSTANT_TYPE_DOUBLE = -10
CONSTANT_TYPE_STRING = -11
CONSTANT_TYPE_BYTE_ARRAY = -12
CONSTANT_TYPE_BOOLEAN_ARRAY = -13
CONSTANT_TYPE_CHAR_ARRAY = -14
CONSTANT_TYPE_SHORT_ARRAY = -15
CONSTANT_TYPE_INTEGER_ARRAY = -16
CONSTANT_TYPE_LONG_ARRAY = -17
CONSTANT_TYPE_FLOAT_ARRAY = -18
CONSTANT_TYPE_DOUBLE_ARRAY = -19
CONSTANT_TYPE_STRING_ARRAY = -20
CONSTANT_TYPE_UUID = -21

# ------------------------------------------------------------
# DEFAULT SERIALIZERS

JAVA_DEFAULT_TYPE_CLASS = -24
JAVA_DEFAULT_TYPE_DATE = -25
JAVA_DEFAULT_TYPE_BIG_INTEGER = -26
JAVA_DEFAULT_TYPE_ARRAY = -28
JAVA_DEFAULT_TYPE_ARRAY_LIST = -29
JAVA_DEFAULT_TYPE_LINKED_LIST = -30
JAVASCRIPT_JSON_SERIALIZATION_TYPE = -130

# ------------------------------------------------------------
# JAVA SERIALIZATION

JAVA_DEFAULT_TYPE_SERIALIZABLE = -100
PYTHON_TYPE_PICKLE = -120
